import logging
from typing import Optional, Dict

from meross_iot.model.enums import Namespace, RollerShutterState, DNDMode

_LOGGER = logging.getLogger(__name__)


class SystemDndMixin:
    _execute_command: callable
    check_full_update_done: callable

    ## It looks like the DND mode update/change does not trigger any PUSH notification update.

    async def async_get_dnd_mode(self, timeout: Optional[float] = None, *args, **kwargs) -> DNDMode:
        """
        Polls the device and retrieves its DO-NOT-DISTURB mode.
        This method will actually refresh the cached DNDMode by issuing a MQTT message to the broker.
        You should avoid using this method when not strictly needed and rely on the cached DNDMode available
        via `get_dnd_mode()`.
        :param timeout:
        :param args:
        :param kwargs:
        :return:
        """
        result = await self._execute_command(method="GET",
                                             namespace=Namespace.SYSTEM_DND_MODE,
                                             payload={},
                                             timeout=timeout)
        res = DNDMode(result['DNDMode']['mode'])
        return res

    async def set_dnd_mode(self, mode: DNDMode, timeout: Optional[float] = None, *args, **kwargs) -> None:
        """
        Controls the DND Mode setting on this device. When "Do not disturb" mode is enabled,
        the device will turn off its ambient led.
        :param mode:
        :param timeout:
        :param args:
        :param kwargs:
        :return:
        """
        await self._execute_command(method="SET",
                                             namespace=Namespace.SYSTEM_DND_MODE,
                                             payload={'DNDMode': {'mode': mode.value}},
                                             timeout=timeout)



import logging
from typing import Optional, Dict

from meross_iot.model.enums import Namespace, RollerShutterState

_LOGGER = logging.getLogger(__name__)


class RollerShutterTimerMixin:
    _execute_command: callable
    check_full_update_done: callable
    uuid: str
    _shutter__state_by_channel: Dict[int, RollerShutterState]
    _shutter__position_by_channel: Dict[int, int]
    _shutter__config_by_channel: Dict[int, Dict]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self._shutter__state_by_channel = {}
        self._shutter__position_by_channel = {}
        self._shutter__config_by_channel = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.ROLLER_SHUTTER_STATE:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('state')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'state' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The roller shutter timer state push notification contains an object for every channel handled by the
                # device
                for roller_shutter in payload:
                    channel_index = roller_shutter['channel']
                    state = RollerShutterState(roller_shutter['state']) # open (position=100, state=1), close (position=0, state=2), stop (position=-1, state=0)
                    self._shutter__state_by_channel[channel_index] = state
                    locally_handled = True
        elif namespace == Namespace.ROLLER_SHUTTER_POSITION:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('position')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'position' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The roller shutter timer position push notification contains an object for every channel handled by the
                # device
                for roller_shutter in payload:
                    channel_index = roller_shutter['channel']
                    position = roller_shutter['position'] # open (position=100, state=1), close (position=0, state=2), stop (position=-1, state=0)
                    self._shutter__position_by_channel[channel_index] = position
                    locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_open(self, channel: int = 0, *args, **kwargs) -> None:
        """
        Operates the roller shutter: sends the open command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(position=100, channel=channel, *args, **kwargs)

    async def async_stop(self, channel: int = 0, *args, **kwargs) -> None:
        """
        Operates the roller shutter: sends the stop command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(position=-1, channel=channel, *args, **kwargs)

    async def async_close(self, channel: int = 0, *args, **kwargs) -> None:
        """
        Operates the roller shutter: sends the close command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(position=0, channel=channel, *args, **kwargs)

    async def _async_operate(self, position: int, channel: int = 0, timeout: Optional[float] = None, *args, **kwargs) -> None:
        payload = {'position': {"position": position, "channel": channel}}
        await self._execute_command(method="SET",
                                    namespace=Namespace.ROLLER_SHUTTER_POSITION,
                                    payload=payload,
                                    timeout=timeout)
        # Respect to other devices, we don't assume the command was immediately successful as the
        # state/position change might take some time to take place.
        # self.__state_by_channel[channel] = state
        # self._roller_shutter_position_by_channel[channel] = position

    async def async_update(self, timeout: Optional[float] = None, *args, **kwargs) -> None:
        # Call the super implementation
        await super().async_update(*args, **kwargs)
        # Update the configuration at the same time
        await self.async_fetch_config()

    async def async_fetch_config(self, timeout: Optional[float] = None, *args, **kwargs) -> None:
        data = await self._execute_command(method="GET",
                                    namespace=Namespace.ROLLER_SHUTTER_CONFIG,
                                    payload={},
                                    timeout=timeout)
        config = data.get('config')
        for d in config:
            channel = d['channel']
            channel_config = d.copy()
            self._shutter__config_by_channel[channel] = channel_config

    def get_open_timer_duration_millis(self, channel: int = 0, *args, **kwargs) -> int:
        self.check_full_update_done()
        return self._shutter__config_by_channel.get(channel).get("signalOpen")

    def get_close_timer_duration_millis(self, channel: int = 0, *args, **kwargs) -> int:
        self.check_full_update_done()
        return self._shutter__config_by_channel.get(channel).get("signalClose")

    async def async_set_config(self, open_timer_seconds: int, close_timer_seconds: int, channel: int = 0, timeout: Optional[float] = None, *args, **kwargs) -> None:
        """
        Sets the configuration parameters for the roller shutter on the given channel.
        :param open_timer_seconds: open timer, min 10, max 120.
        :param close_timer_seconds: close timer, min 10, max 120.
        :param channel: channel to configure
        :param timeout:
        :return:
        """
        if open_timer_seconds < 10 or open_timer_seconds > 120:
            raise ValueError("Invalid open_timer_seconds timer, must be between 10 and 120 seconds.")
        if close_timer_seconds < 10 or close_timer_seconds > 120:
            raise ValueError("Invalid close_timer_seconds timer, must be between 10 and 120 seconds.")

        config = {"channel": channel, "signalOpen": 1000*open_timer_seconds, "signalClose": 1000*close_timer_seconds}
        res = await self._execute_command(method="SET",
                                    namespace=Namespace.ROLLER_SHUTTER_CONFIG,
                                    payload={"config": config},
                                    timeout=timeout)
        self._shutter__config_by_channel[channel] = config

    def get_status(self, channel: int = 0, *args, **kwargs) -> RollerShutterState:
        """
        The current roller shutter status. Returns 1 if the given roller shutter is open, 2 if it is close, 0 if it is stop.

        :param channel: channel of which status is needed

        :return: 1 if the given roller shutter is opened, 2 if it is closed, 0 if it is stopped.
        """
        self.check_full_update_done()
        state = self._shutter__state_by_channel.get(channel)
        return RollerShutterState(state) if state is not None else RollerShutterState.UNKNOWN

    def get_position(self, channel: int = 0, *args, **kwargs) -> Optional[int]:
        """
        The current roller shutter position. Returns 100 if the given roller shutter is open, 0 if it is close, -1 if it is stop.

        :param channel: channel of which status is needed

        :return: 100 if the given roller shutter is opened, 0 if it is closed, -1 if it is stopped.
        """
        self.check_full_update_done()
        return self._shutter__position_by_channel.get(channel)


import asyncio
import getpass
import json
import logging
import os
import ssl
import sys
import uuid as UUID
from hashlib import md5
from os import path, environ
from threading import Event
from zipfile import ZipFile
import paho.mqtt.client as mqtt
from meross_iot.http_api import MerossHttpClient
from meross_iot.manager import MerossManager
from meross_iot.model.constants import DEFAULT_MQTT_PORT
from meross_iot.model.enums import Namespace, OnlineStatus
from meross_iot.utilities.mqtt import build_device_request_topic, build_client_response_topic, build_client_user_topic
from meross_iot.utilities.network import extract_domain, extract_port

SNIFF_LOG_FILE = 'sniff.log'
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
l = logging.getLogger().getChild("Sniffer")
l.setLevel(logging.DEBUG)
lhandler = logging.FileHandler(mode='w', filename=SNIFF_LOG_FILE)
lhandler.setFormatter(formatter)
l.addHandler(lhandler)

ROOT_LOG_FILE = 'root.log'
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
r = logging.getLogger()
r.setLevel(logging.DEBUG)
rhandler = logging.FileHandler(mode='w', filename=ROOT_LOG_FILE)
rhandler.setFormatter(formatter)
r.addHandler(rhandler)
r.setLevel(logging.DEBUG)


class DeviceSniffer(object):
    def __init__(self, user_id, hashed_password, target_device_uuid, ca_cert=None, mqtt_host="iot.meross.com", mqtt_port=2001):
        self.connect_event = Event()
        self.subscribe_event = Event()
        self.user_id = user_id

        self.device_topic = None
        self.client_response_topic = None
        self.user_topic = None

        self.mqtt_domain = mqtt_host
        self.mqtt_port = mqtt_port
        self.target_device_uuid = target_device_uuid

        # Generate random app and client id
        md5_hash = md5()
        rnd_uuid = UUID.uuid4()
        md5_hash.update(("%s%s" % ("API", rnd_uuid)).encode("utf8"))
        self._app_id = "sniffer"
        self._client_id = 'app:sniffer-%s' % md5_hash.hexdigest()

        self._mqtt_client = mqtt.Client(client_id=self._client_id,
                                        protocol=mqtt.MQTTv311)

        self._mqtt_client.on_connect = self._on_connect
        self._mqtt_client.on_message = self._on_message
        self._mqtt_client.on_disconnect = self._on_disconnect
        self._mqtt_client.on_subscribe = self._on_subscribe

        # Avoid login if user_id is None
        if user_id is not None:
            self._mqtt_client.username_pw_set(username=user_id,
                                              password=hashed_password)
        self._mqtt_client.tls_set(ca_certs=ca_cert, certfile=None,
                                  keyfile=None, cert_reqs=ssl.CERT_REQUIRED,
                                  tls_version=ssl.PROTOCOL_TLS_CLIENT,
                                  ciphers=None)

    def _on_connect(self, client, userdata, rc, other):
        l.debug("Connected to MQTT Broker")
        self.connect_event.set()

    def start(self):
        """
        Starts the connection to the MQTT broker
        :return:
        """
        l.debug("Initializing the MQTT connection...")
        self._mqtt_client.connect(self.mqtt_domain, self.mqtt_port, keepalive=30)

        # Starts a new thread that handles mqtt protocol and calls us back via callbacks
        l.debug("(Re)Starting the MQTT loop.")
        self._mqtt_client.loop_stop(True)
        self._mqtt_client.loop_start()
        self.connect_event.wait()

        # Subscribe to the corresponding topics ...
        self.device_topic = build_device_request_topic(self.target_device_uuid)
        self.client_response_topic = build_client_response_topic(self.user_id, self._app_id)
        self.user_topic = build_client_user_topic(self.user_id)

        l.info(f"Subscribing to topic: {self.device_topic}")
        self._mqtt_client.subscribe(self.device_topic)
        self.subscribe_event.wait()
        self.subscribe_event.clear()

        l.info(f"Subscribing to topic: {self.client_response_topic}")
        self._mqtt_client.subscribe(self.client_response_topic)
        self.subscribe_event.wait()
        self.subscribe_event.clear()

        l.info(f"Subscribing to topic: {self.user_topic}")
        self._mqtt_client.subscribe(self.user_topic)
        self.subscribe_event.wait()
        self.subscribe_event.clear()

    def stop(self):
        self._mqtt_client.disconnect()

    def _on_subscribe(self, client, userdata, mid, granted_qos):
        l.debug("Subscribed correctly")
        self.subscribe_event.set()

    def _on_message(self, client, userdata, msg):
        message = json.loads(str(msg.payload, "utf8"))
        header = message['header']

        topic_str = "Unknown"
        if msg.topic == self.user_topic:
            topic_str = "USER-TOPIC"
        elif msg.topic == self.client_response_topic:
            topic_str = "CLIENT-RESPONSE-TOPIC"
        elif msg.topic == self.device_topic:
            topic_str = "DEVICE-TOPIC"

        l.info("%s (%s) <- %s" % (topic_str, msg.topic, message))

    def _on_disconnect(self, client, userdata, rc):
        l.debug("Disconnected from MQTT brocker")


async def _main():
    print("Welcome to the Sniffer utility. This python script will gather some useful information about your "
          "Meross devices. All the collected information will be zipped into a zip archive. "
          "You could share such zip file with the developers to help them add support for your device. "
          "Although this utility won't collect your email/password, we recommend you to change "
          "your Meross account password to a temporary one before using this software. Once you are done, "
          "you can restore back your original password. By doing so, you are 100% sure you are not leaking any "
          "real password to the developers.")
    email = environ.get("MEROSS_EMAIL")
    if email is None:
        email = input("Please specify your meross email: ")
        email = email.strip()

    password = environ.get("MEROSS_PASSWORD")
    if password is None:
        password = getpass.getpass(prompt='Please specify your Meross password: ', stream=sys.stdout).strip()

    devices = []
    http = None

    # Gather HTTP devices
    try:
        http = await MerossHttpClient.async_from_user_password(email, password)
        print("# Collecting devices via HTTP api...")
        devices = await http.async_list_devices()
    except:
        print("An error occurred while retrieving Meross devices.")
        exit(1)

    for i, d in enumerate(devices):
        print(f"[{i}] - {d.dev_name} ({d.device_type}) - {d.online_status.name}")

    print("Please note that some devices rely on a HUB. Those devices won't be listed: in such cases, you need"
          "to select the corresponding HUB device.")

    while True:
        selection = input("Select the device you want to study (numeric index): ")
        selection = int(selection.strip())
        selected_device = devices[selection]
        if selected_device is not None:
            break

    print(f"You have selected {selected_device.dev_name}.")
    if selected_device.online_status != OnlineStatus.ONLINE:
        print("!! WARNING !! You selected a device that has not been reported as online. ")

    # Start the manager
    creds = http.cloud_credentials
    md5_hash = md5()
    clearpwd = "%s%s" % (creds.user_id, creds.key)
    md5_hash.update(clearpwd.encode("utf8"))
    hashed_password = md5_hash.hexdigest()
    sniffer = DeviceSniffer(
        creds.user_id,
        hashed_password,
        selected_device.uuid,
        ca_cert=None,
        mqtt_host=extract_domain(selected_device.domain) or "iot.meross.com"
    )

    print("Starting the sniffer...")
    sniffer.start()
    print("You can now start commanding this device from the Meross APP. This utility will catch "
          "every command you send from the app to this device. Such data could help developers to "
          "reproduce the functionality on this library. When DONE, press ENTER to finish.")

    input("Press ENTER to finish.")
    sniffer.stop()

    # As very last step, try to collect data via get_all() and get_abilities
    l.info("--------------- More data -----------------")
    print("Collecting state info...")
    manager = None
    try:
        manager = MerossManager(http_client=http)
        await manager.async_init()

        mqtt_host = selected_device.get_mqtt_host()
        mqtt_port = selected_device.get_mqtt_port()

        # Manually get device abilities
        response_all = await manager.async_execute_cmd(destination_device_uuid=selected_device.uuid,
                                                       method="GET",
                                                       namespace=Namespace.SYSTEM_ALL,
                                                       payload={},
                                                       mqtt_hostname=mqtt_host,
                                                       mqtt_port=mqtt_port)
        response_abilities = await manager.async_execute_cmd(destination_device_uuid=selected_device.uuid,
                                                             method="GET",
                                                             namespace=Namespace.SYSTEM_ABILITY,
                                                             payload={},
                                                             mqtt_hostname=mqtt_host,
                                                             mqtt_port=mqtt_port
                                                             )

        l.info(f"Sysdata for {selected_device.dev_name} ({selected_device.uuid}): {response_all}")
        l.info(f"Abilities for {selected_device.dev_name} ({selected_device.uuid}): {response_abilities}")
    except:
        l.exception(f"Could not collect sysdata/abilities for {selected_device.uuid}")
    finally:
        if manager is not None:
            manager.close()
        await http.async_logout()

    print("Collecting logs...")
    zipObj = ZipFile('data.zip', 'w')
    zipObj.write(SNIFF_LOG_FILE)
    zipObj.write(ROOT_LOG_FILE)
    zipObj.close()

    print("A zipfile has been created containing the logs collected during this execution. "
          "It is located in {path}.".format(path=path.abspath(zipObj.filename)))
    print("Thanks for helping the Meross community!")


def main():
    # On Windows + Python 3.8, you should uncomment the following
    if os.name == 'nt':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    loop = asyncio.get_event_loop()
    loop.run_until_complete(_main())
    loop.close()


if __name__ == '__main__':
    main()



import time
from collections import deque
from datetime import timedelta
from typing import Optional, Deque, Dict, ItemsView

from meross_iot.http_api import ErrorCodes


class HttpRequestSample:
    """
    Helper class to hold a single request the http API server
    """
    def __init__(self,
                 method: str,
                 url: str,
                 http_response_code: int,
                 api_response_code: ErrorCodes,
                 timestamp: Optional[float] = None):

        self._method = method
        self._url = url
        self._http_response_code = http_response_code
        self._api_response_code = api_response_code
        self._ts = timestamp

    @property
    def method(self):
        """
        HTTP Method
        :return:
        """
        return self._method

    @property
    def url(self):
        """
        URL
        :return:
        """
        return self._url

    @property
    def http_response_code(self):
        """
        HTTP response code
        :return: 
        """
        return self._http_response_code

    @property
    def api_response_code(self) -> ErrorCodes:
        """
        Meross HTTP API response code
        :return:
        """
        return self._api_response_code

    @property
    def timestamp(self):
        """
        Timestamp of the request
        :return:
        """
        return self._ts


class ApiCallSample:
    """
    Helper class to hold a single call to the mqtt broker
    """
    def __init__(self,
                 device_uuid: str,
                 namespace: str,
                 method: str,
                 timestamp: Optional[float] = None):

        self._uuid = device_uuid
        self._ns = namespace
        self._method = method
        self._ts = timestamp

    @property
    def device_uuid(self):
        """
        Target Device UUID for the message
        """
        return self._uuid

    @property
    def namespace(self):
        """
        Target Namespace for the message
        """
        return self._ns

    @property
    def method(self):
        """
        Target Method for the message
        """
        return self._method

    @property
    def timestamp(self):
        """
        When the record has been registered
        """
        return self._ts


class HttpStat:
    """
    Helper class that handles HTTP stats result for a single URL
    """
    def __init__(self):
        self._total_api_calls = 0
        self._by_http_response_code: Dict[int, int] = {}
        self._by_api_response_code: Dict[ErrorCodes, int] = {}

    def add(self, sample: HttpRequestSample) -> None:
        self._total_api_calls += 1

        # Aggregate by HTTP response code
        if sample.http_response_code not in self._by_http_response_code:
            self._by_http_response_code[sample.http_response_code] = 1
        else:
            self._by_http_response_code[sample.http_response_code] += 1

        # Aggregate by API response code
        if sample.api_response_code not in self._by_api_response_code:
            self._by_api_response_code[sample.api_response_code] = 1
        else:
            self._by_api_response_code[sample.api_response_code] += 1

    @property
    def total_calls(self) -> int:
        """
        Total number of calls for this URL
        """
        return self._total_api_calls

    def by_http_reponse_code(self) -> ItemsView[int, int]:
        """
        Number of calls for this url aggregated by HTTP response code
        """
        return self._by_http_response_code.items()

    def by_api_status_code(self) -> ItemsView[ErrorCodes, int]:
        """
        Number of calls for this url aggregated by API response status code
        """
        return self._by_api_response_code.items()

    def __repr__(self):
        http_reponses = sorted(self._by_http_response_code.items(), key=lambda item: item[1], reverse=True)
        http_details = ", ".join([f"{code}: {calls}" for code, calls in http_reponses])
        api_statuses = sorted(self._by_http_response_code.items(), key=lambda item: item[1], reverse=True)
        api_details = ", ".join([f"{code}: {calls}" for code, calls in api_statuses])
        return f"{self.total_calls} ({http_details}; {api_details})"


class ApiStat:
    """
    Helper class that handles API stats result for a single device
    """
    def __init__(self):
        self._total_api_calls = 0
        self._by_method_namespace: Dict[str, int] = {}

    def add(self, sample: ApiCallSample) -> None:
        self._total_api_calls += 1

        method_ns = f"{sample.method} {sample.namespace}"
        if method_ns not in self._by_method_namespace:
            self._by_method_namespace[method_ns] = 1
        else:
            self._by_method_namespace[method_ns] += 1

    @property
    def total_calls(self) -> int:
        """
        Total number of calls for this device
        """
        return self._total_api_calls

    def by_method_namespace(self) -> ItemsView[str, int]:
        """
        Number of calls for this device aggregated by Method/Namespace
        """
        return self._by_method_namespace.items()

    def __repr__(self):
        top_calls = sorted(self._by_method_namespace.items(), key=lambda item: item[1], reverse=True)
        details = ", ".join([f"{method}: {calls}" for method, calls in top_calls])
        return f"{self.total_calls} ({details})"


class HttpStatsResult:
    """
    Helper class that handles HTTP requests stats result
    """
    def __init__(self):
        self._global = HttpStat()
        self._by_url: Dict[str, HttpStat] = {}

    def add(self, sample: HttpRequestSample):
        self._global.add(sample)
        byurl = self._by_url.get(sample.url)
        if byurl is None:
            byurl = HttpStat()
            self._by_url[sample.url] = byurl
        byurl.add(sample)

    @property
    def global_stats(self) -> HttpStat:
        """
        Total number of calls
        """
        return self._global

    def stats_by_url(self, url: str) -> Optional[HttpStat]:
        """
        Returns the statistics of a specific url
        """
        return self._by_url.get(url)

    def device_stats(self) -> ItemsView[str, HttpStat]:
        """
        Statistics aggregated by URL
        """
        return self._by_url.items()

    def __repr__(self):
        top_urls = sorted(self._by_url.items(), key=lambda item: item[1].global_stats, reverse=True)
        url_rerp = ",\n".join([f"{url}: {stats}" for url, stats in top_urls])
        return f"--------\n" \
               f"Global Calls: {self._global.total_calls}\n" \
               f"{url_rerp}\n" \
               f"--------\n"


class ApiStatsResult:
    """
    Helper class that handles API stats result
    """
    def __init__(self):
        self._global = ApiStat()
        self._by_uuid: Dict[str, ApiStat] = {}

    def add(self, sample: ApiCallSample):
        self._global.add(sample)
        byuuid = self._by_uuid.get(sample.device_uuid)
        if byuuid is None:
            byuuid = ApiStat()
            self._by_uuid[sample.device_uuid] = byuuid
        byuuid.add(sample)

    @property
    def global_stats(self) -> ApiStat:
        """
        Total number of calls
        """
        return self._global

    def stats_by_uuid(self, device_uuid) -> Optional[ApiStat]:
        """
        Returns the statistics of a specific device UUID
        """
        return self._by_uuid.get(device_uuid)

    def device_stats(self) -> ItemsView[str, ApiStat]:
        """
        Statistics aggregated by Device UUID
        """
        return self._by_uuid.items()

    def __repr__(self):
        top_devices = sorted(self._by_uuid.items(), key=lambda item: item[1].global_stats, reverse=True)
        device_rerp = ",\n".join([f"{uuid}: {stats}" for uuid, stats in top_devices])
        return f"--------\n" \
               f"Global Calls: {self._global.total_calls}\n" \
               f"{device_rerp}\n" \
               f"--------\n"


class HttpStatsCounter:
    """
    Helper class to keep track and calculate statistics for sent HTTP requests
    """
    def __init__(self, max_samples=1000):
        self._samples: Deque[HttpRequestSample] = deque([], maxlen=max_samples)

    def notify_http_request(self, request_url: str, method: str, http_response_code: int, api_response_code: Optional[ErrorCodes]):
        sample = HttpRequestSample(
            url=request_url,
            method=method,
            http_response_code=http_response_code,
            api_response_code=api_response_code,
            timestamp=time.time()
        )
        self._samples.append(sample)

    def _get_stats(self, samples: Deque[HttpRequestSample], time_window: timedelta = timedelta(minutes=1)) -> HttpStatsResult:
        result = HttpStatsResult()
        lower_limit = time.time() - time_window.total_seconds()
        for sample in reversed(samples):
            if sample.timestamp > lower_limit:
                result.add(sample)

        return result

    def get_stats(self, time_window: timedelta = timedelta(minutes=1)) -> HttpStatsResult:
        """
        Returns the statistics of sent MQTT messages to the MQTT broker
        """
        return self._get_stats(samples=self._samples, time_window=time_window)


class ApiCounter:
    """
    Helper class to keep track and calculate statistics for sent MQTT message
    """
    def __init__(self, max_samples=1000):
        self.api_calls: Deque[ApiCallSample] = deque([], maxlen=max_samples)
        self.delayed_calls: Deque[ApiCallSample] = deque([], maxlen=max_samples)
        self.dropped_calls: Deque[ApiCallSample] = deque([], maxlen=max_samples)

    def notify_api_call(self, device_uuid: str, namespace: str, method: str):
        """
        Method called internally by the manager itself, whenever a message is sent to the
        MQTT broker.
        """
        sample = ApiCallSample(
            device_uuid=device_uuid,
            namespace=namespace,
            method=method,
            timestamp=time.time()
        )
        self.api_calls.append(sample)

    def notify_delayed_call(self, device_uuid: str, namespace: str, method: str):
        """
        Method called internally by the manager itself, whenever a message is delayed instead of being sent to the
        MQTT broker.
        """
        sample = ApiCallSample(
            device_uuid=device_uuid,
            namespace=namespace,
            method=method,
            timestamp=time.time()
        )
        self.delayed_calls.append(sample)

    def notify_dropped_call(self, device_uuid: str, namespace: str, method: str):
        """
        Method called internally by the manager itself, whenever a message is dropped instead of being sent to the
        MQTT broker.
        """
        sample = ApiCallSample(
            device_uuid=device_uuid,
            namespace=namespace,
            method=method,
            timestamp=time.time()
        )
        self.dropped_calls.append(sample)

    def _get_stats(self, api_samples: Deque[ApiCallSample], time_window: timedelta = timedelta(minutes=1)) -> ApiStatsResult:
        result = ApiStatsResult()
        lower_limit = time.time() - time_window.total_seconds()
        for sample in reversed(api_samples):
            if sample.timestamp > lower_limit:
                result.add(sample)

        return result

    def get_api_stats(self, time_window: timedelta = timedelta(minutes=1)) -> ApiStatsResult:
        """
        Returns the statistics of sent MQTT messages to the MQTT broker
        """
        return self._get_stats(api_samples=self.api_calls, time_window=time_window)

    def get_delayed_api_stats(self, time_window: timedelta = timedelta(minutes=1)) -> ApiStatsResult:
        """
        Returns the statistics of delayed MQTT messages to the MQTT broker
        """
        return self._get_stats(api_samples=self.delayed_calls, time_window=time_window)

    def get_dropped_api_stats(self, time_window: timedelta = timedelta(minutes=1)) -> ApiStatsResult:
        """
        Returns the statistics of dropped MQTT messages to the MQTT broker
        """
        return self._get_stats(api_samples=self.dropped_calls, time_window=time_window)




import json
from datetime import datetime
from typing import Union


class MerossCloudCreds(object):
    """
    Utility class that holds Meross Cloud credentials for a given user.
    """
    def __init__(self, token: str, key: str, user_id: str, user_email: str, issued_on: Union[datetime, str]):
        self.token = token
        self.key = key
        self.user_id = user_id
        self.user_email = user_email

        if isinstance(issued_on, datetime):
            self.issued_on = issued_on
        else:
            self.issued_on = datetime.fromisoformat(issued_on)

    def to_json(self) -> str:
        """
        Serializes the current object to a json string
        :return:
        """
        return json.dumps({
            "token": self.token,
            "key": self.key,
            "user_id": self.user_id,
            "user_email": self.user_email,
            "issued_on": self.issued_on.isoformat()
        })

    @classmethod
    def from_json(cls, json_string: str):
        """
        Builds a MerossCloudCreds instance from a previously serialized json string
        :param json_string:
        :return:
        """
        data = json.loads(json_string)
        return MerossCloudCreds(**data)

    def __repr__(self):
        return self.to_json()



import pkg_resources

from meross_iot import name


def current_version():
    try:
        return pkg_resources.get_distribution(name).version
    except pkg_resources.DistributionNotFound:
        return "0.0.0"



from datetime import datetime, timedelta

class ErrorBudget:
    def __init__(self, initial_budget:int, window_start:datetime):
        self.budget = initial_budget
        self.window_start = window_start

class ErrorBudgetManager:

    def __init__(self, max_errors:int = 1, time_window: timedelta = timedelta(seconds=60)):
        self._devices_budget={}
        self._window = time_window
        self._max_errors = max_errors

    def _get_update_budget_window(self, device_uuid: str) -> ErrorBudget:
        dev_budget = self._devices_budget.get(device_uuid)
        if dev_budget is None:
            dev_budget = ErrorBudget(self._max_errors, datetime.utcnow())

        # Re-init the error budget if window expired
        if datetime.utcnow() > (dev_budget.window_start + self._window):
            dev_budget.budget = self._max_errors
            dev_budget.window_start = datetime.utcnow()

        return dev_budget

    def notify_error(self, device_uuid: str):
        dev_budget = self._get_update_budget_window(device_uuid)
        if dev_budget.budget < 1:
            return
        else:
            dev_budget.budget -= 1

    def is_out_of_budget(self, device_uuid: str) -> bool:
        budget = self._get_update_budget_window(device_uuid)
        return budget.budget < 1

